#
# cache.py - Cache of BBS.
# Copyright (c) 2005 shinGETsu Project.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# $Id: cache.py,v 1.35 2005/06/20 07:25:14 fuktommy Exp $
#

import base64
import md5
import os
import random
import re
import shutil
import sys
import time

import apollo
import config
from node import *

__version__ = "$Revision: 1.35 $"
__all__ = ["Record", "Cache", "CacheList", "UpdateList"]

class Record(dict):

    """One article of BBS.

    timestamp<>id<>foo:bar<>hoge:fuga<>...
    """

    recstr = ""     # record in string format
    datfile = ""    # cache file
    stamp = 0       # date when the record made
    id = ""         # md5sum
    idstr = ""      # stamp_id
    path = ""       # path for real file
    rm_path = ""    # path for removed marker

    def __init__(self, datfile="", idstr=""):
        dict.__init__(self)
        self.datfile = datfile
        self.idstr = idstr

        if idstr != "":
            buf = idstr.split("_")
            try:
                self.stamp = int(buf[0])
                self.id = buf[1]
            except ValueError:
                sys.stderr.write(buf + ": bad format\n")
                pass
        self.setpath()

    def __str__(self):
        return self.recstr

    def __gt__(self, y):
        if self.stamp != y.stamp:
            return self.stamp > y.stamp
        else:
            return self.idstr > y.idstr

    def __lt__(self, y):
        if self.stamp != y.stamp:
            return self.stamp < y.stamp
        else:
            return self.idstr < y.idstr

    def __eq__(self, y):
        return self.idstr == y.idstr

    def __ne__(self, y):
        return self.idstr != y.idstr

    def setpath(self):
        if (self.idstr != "") and (self.datfile != ""):
            self.path = "/".join((config.cache_dir,
                                  self.datfile,
                                  "record",
                                  self.idstr))
            self.rm_path = "/".join((config.cache_dir,
                                     self.datfile,
                                     "removed",
                                     self.idstr))

    def parse(self, recstr):
        """Parse cache record."""
        self.recstr = re.sub(r"[\r\n]*$", "", recstr)
        tmp = self.recstr.split("<>")
        try:
            self["stamp"] = tmp.pop(0)
            self["id"] = tmp.pop(0)
            self.idstr = self["stamp"] + "_" + self["id"]
            self.stamp = int(self["stamp"])
            self.id = self["id"]
        except (ValueError, KeyError, IndexError):
            sys.stderr.write(recstr + ": bad format\n")
            return False
        for i in tmp:
            buf = i.split(":", 1)
            if len(buf) >= 2:
                buf[1] = re.sub(r"<br>", "\n", buf[1])
                buf[1] = re.sub(r"<", "&lt;", buf[1])
                buf[1] = re.sub(r">", "&gt;", buf[1])
                buf[1] = re.sub(r"\n", "<br>", buf[1])
                self[buf[0]] = buf[1]
        self.setpath()
        return True

    def load(self):
        if self.path == "":
            sys.stderr.write("Null file name\n")
            return False
        try:
            if self.size() <= 0:
                self.remove()
                return False
            f = file(self.path)
            parse_ok = self.parse(f.readline())
            f.close()
            return parse_ok
        except IOError:
            sys.stderr.write(self.path + ": IOError\n")
            return False

    def build(self, stamp, body, passwd=""):
        target = body.keys()
        bodyary = []
        for key in target:
            bodyary.append(key + ":" + body[key])
            self[key] = body[key]
        bodystr = "<>".join(bodyary)
        if passwd != "":
            (pubkey, prikey) = apollo.key_pair(passwd)
            md = md5.new(bodystr).hexdigest()
            sign = apollo.sign(md, pubkey, prikey)
            self["pubkey"] = pubkey
            self["sign"] = sign
            self["target"] = ",".join(target)

            bodystr += "<>pubkey:" + pubkey + \
                       "<>sign:" + sign + \
                       "<>target:" + ",".join(target)

        id = md5.new(bodystr).hexdigest()
        self.stamp = int(stamp)
        self.recstr = str(stamp) + "<>" + id + "<>" + bodystr
        self.idstr = str(stamp) + "_" + id
        self["stamp"] = str(stamp)
        self["id"] = id
        self.id = id
        self.setpath()
        return id

    def md5check(self):
        buf = str(self).split("<>", 2)
        if (len(buf) > 2):
            return md5.new(buf[2]).hexdigest() == self["id"]
        else:
            return False

    def attach_path(self, suffix=None):
        if self.path == "":
            sys.stderr.write("Null file name\n")
            return None
        dir = "/".join((config.cache_dir, self.datfile, "attach"))
        if suffix is not None:
            suffix = re.sub(r"[^-_.A-Za-z0-9]", "", suffix)
            if suffix == "":
                suffix = "txt"
            return dir + "/" + self.idstr + "." + suffix
        for i in os.listdir(dir):
            if i.startswith(self.idstr):
                return dir + "/" + i
        return None

    def attach_size(self, path=None):
        if path is None:
            path = self.attach_path()
        if path is None:
            return 0
        else:
            return os.path.getsize(path)

    def _write_file(self, path, data):
        """Save data to file."""
        if path == "":
            sys.stderr.write("Null file name\n")
            return False
        try:
            f = file(path, "wb")
            f.write(data + "\n")
            f.close()
            return True
        except IOError:
            sys.stderr.write(path + ": IOError\n")
            return False

    def sync(self, force=False):
        """Save files."""
        if self.removed():
            return
        elif (not force) and self.exists():
            pass
        else:
            self._write_file(self.path, str(self))
        if ("suffix" in self) and ("attach" in self):
            self._write_file(self.attach_path(self["suffix"]),
                             base64.decodestring(self["attach"]))
        elif "attach" in self:
            self._write_file(self.attach_path("txt"),
                             base64.decodestring(self["attach"]))

    def body_string(self):
        """Remove attach field."""
        buf = [self["stamp"], self["id"]]
        for k in self:
            if k in ("stamp", "id"):
                pass
            elif k == "attach":
                buf.append("attach:1")
            elif k == "sign":
                pass
            elif k == "pubkey":
                if self.check_sign():
                    short_key = apollo.cut_key(self["pubkey"])
                    buf.append("pubkey:" + short_key)
            else:
                buf.append(k + ":" + self[k])
        return "<>".join(buf)

    def check_sign(self):
        """Check sigunature."""
        for k in ("pubkey", "sign", "target"):
            if k not in self:
                return False
        target = ""
        for t in self["target"].split(","):
            try:
                target += "<>" + t + ":" + self[t]
            except KeyError:
                return False
            except ValueError:
                return False
        target = target[2:]         # remove ^<>

        md = md5.new(target).hexdigest()
        if apollo.verify(md, self["sign"], self["pubkey"]):
            return True
        else:
            return False

    def exists(self):
        return os.path.isfile(self.path)

    def removed(self):
        return os.path.isfile(self.rm_path)

    def size(self):
        try:
            return os.path.getsize(self.path)
        except OSError, (errno, strerror):
            sys.stderr.write("%s: OSError: %s\n" % (self.path, strerror))
            return 0

    def remove(self):
        """Remove record.

        The record is removed from cache,
        but do not removed from file-system.
        """
        try:
            shutil.move(self.path, self.rm_path)
            apath = self.attach_path()
            if (apath is not None) and os.path.isfile(apath):
                os.remove(apath)
            return True
        except (IOError, OSError), (errorno, strerror):
            sys.stderr.write("%s: IOError/OSError: %s\n"%(self.path, strerror))
            return False

# End of Record


class Cache(dict):

    """Cache of BBS.

    Plain text (encode: UTF-8).
    One record par one line.
    """

    datfile = ""
    datpath = config.cache_dir
    stamp = 0       # when the cache is modified
    size = 0        # size of cache file
    type = ""       # list/thread/note

    def __init__(self, datfile):
        dict.__init__(self)
        self.datfile = datfile
        self.datpath += "/" + datfile

        self.stamp = self._load_status("stamp")
        self.size = self._load_status("size")

        for type in config.types:
            if self.datfile.startswith(type):
                self.type = type
                break

        if self.exists():
            try:
                for k in os.listdir(self.datpath + "/record"):
                    self[k] = Record(datfile=self.datfile, idstr=k)
            except OSError:
                sys.stderr.write(self.datpath + "/record: OSError\n")

        self.limit_rec = config.limit_rec.get(self.type, 0)
        if (self.limit_rec <= 0) or (len(self) > self.limit_rec):
            self.save_record = config.save_record.get(self.type, 0)
            self.get_range = config.get_range.get(self.type, 0)
            self.sync_range = config.sync_range.get(self.type, 0)
            self.save_removed = config.save_removed.get(self.type, 0)
        else:
            self.save_record = 0
            self.get_range = 0
            self.sync_range = 0
            self.save_removed = 0

    def __str__(self):
        return self.datfile

    def keys(self):
        k = dict.keys(self)
        k.sort()
        return k

    def __iter__(self):
        return iter(self.keys())

    def exists(self):
        return os.path.isdir(self.datpath)

    def _load_status(self, key):
        path = self.datpath + "/" + key + ".stat"
        try:
            f = file(path)
            v = f.readline()
            f.close()
            return int(v.strip())
        except IOError:
            #sys.stderr.write(path + ": IOError\n")
            return 0
        except ValueError:
            sys.stderr.write(path + ": ValueError\n")
            return 0

    def _save_status(self, key, val):
        path = self.datpath + "/" + key + ".stat"
        try:
            f = file(path, "wb")
            f.write(str(val)+"\n")
            f.close()
            return True
        except IOError:
            sys.stderr.write(path + ": IOError\n")
            return False

    def _sync_status(self):
        self._save_status("stamp", self.stamp)
        self._save_status("size", self.size)

    def add_data(self, newdata):
        """Add new data cache."""
        for d in ("", "/attach", "/record", "/removed"):
            if not os.path.isdir(self.datpath + d):
                try:
                    os.makedirs(self.datpath + d)
                except (IOError, OSError):
                    sys.stderr.write(self.datfile + ": IOError/OSError\n")
        body = file(self.datpath + "/body.cache", "ab")
        for r in newdata:
            r.sync()
            self[r.idstr] = r
            body.write(r.body_string() + "\n")
            if self.stamp < r.stamp:
                self.stamp = r.stamp
            self.size += len(str(r)) + 1
        body.close()
        self._sync_status()

    def get_data(self, stamp=0, id="", node=None):
        """Get appointed data."""
        buf = node.talk("/get/" + self.datfile + "/" + str(stamp) + "/" + id)
        flag = False
        for i in buf:
            r = Record(datfile=self.datfile)
            parse_ok = r.parse(i)
            if parse_ok and (r["stamp"] == str(stamp)) and r.md5check():
                self.add_data((r,))
                flag = True
        return flag

    def get_with_range(self, node=None):
        """Get data in range."""
        if self.stamp > 0:
            begin = self.stamp
        else:
            begin = 0
        if self.sync_range > 0:
            now = int(time.time()) - self.sync_range
        else:
            now = 0
        if begin < 0:
            begin = 0
        elif now < begin:
            begin = now

        if (begin <= 0) and (len(self) <= 0):
            if self.get_range > 0:
                begin = int(time.time()) - self.get_range
                if begin < 0:
                    begin = 0
            else:
                begin = 0
            buf = node.talk("/get/%s/%d-" % (self.datfile, begin))
        else:
            buf = node.talk("/head/%s/%d-" % (self.datfile, begin))
            tmp = []
            for head in buf:
                r = Record(datfile=self.datfile,
                           idstr=re.sub(r"<>", "_", head).strip())
                if not (r.exists() or r.removed()):
                    tmp.append(r)
            buf = []
            for r in tmp:
                body = node.talk("/".join(("",
                                           "get",
                                           self.datfile,
                                           str(r.stamp),
                                           r.id)))
                buf.extend(body)

        flag = False
        newdata = []
        for i in buf:
            r = Record(datfile=self.datfile)
            parse_ok = r.parse(i)
            if parse_ok and (begin <= r.stamp) and r.md5check():
                newdata.append(r)
                flag = True
        self.add_data(newdata)
        return flag

    def sync_body(self):
        """Update body.cache."""
        try:
            self.size = 0
            # self.stamp = 0
            body = file(self.datpath + "/body.cache", "wb")
            for i in self:
                r = Record(datfile=self.datfile, idstr=i)
                load_ok = r.load()
                if not load_ok:
                    continue
                r.sync()
                body.write(r.body_string() + "\n")
                if self.stamp < r.stamp:
                    self.stamp = r.stamp
                self.size += len(str(r))
            body.close()
            self._sync_status()
        except IOError:
            sys.stderr.write(self.datpath + "/body.cache: IOError\n")

    def sync_body_light(self, size=0):
        """Update body.cache.
        
        Remove removed records in body.cache.

        """
        try:
            self.size -= size
            body = file(self.datpath + "/body.cache")
            recs = body.readlines()
            body.close()
            body = file(self.datpath + "/body.cache", "wb")
            for line in recs:
                r = Record(datfile=self.datfile)
                parse_ok = r.parse(line)
                if parse_ok and r.exists():
                    body.write(line)
            body.close()
            self._sync_status()
        except IOError:
            sys.stderr.write(self.datpath + "/body.cache: IOError\n")

    def check_attach(self):
        """Remove attach cache that is a field of removed record."""
        try:
            dir = "/".join((config.cache_dir, self.datfile, "attach"))
            for f in os.listdir(dir):
                idstr = f
                i = f.find(".")
                if i >= 0:
                    idstr = f[:i]
                rec = Record(datfile=self.datfile, idstr=idstr)
                if not rec.exists():
                    try:
                        os.remove(dir + "/" + f)
                    except OSError, (errno, strerror):
                        sys.stderr.write("%s/%s: OSError: %s\n" %
                                         (dir, f, strerror))
        except (IOError, OSError), (errno, strerror):
            sys.stderr.write("%s: IOError/OSError: %s\n" %
                             (self.datfile, strerror))

    def load_body(self):
        """Load records from body.cache."""
        try:
            tmp = []
            body = file(self.datpath + "/body.cache")
            for line in body:
                r = Record(datfile=self.datfile)
                if r.parse(line):
                    tmp.append(r)
            body.close()
            return tmp
        except IOError:
            sys.stderr.write(self.datpath + "/body.cache: IOError\n")
            return []

    def remove(self):
        """Remove cache (a.k.a DATFILE).

        It is removed from disk.
        """
        try:
            shutil.rmtree(self.datpath)
            return True
        except IOError, strerror:
            sys.stderr.write("IOError: " + strerror + "\n")
            return False

    def remove_records(self, now, limit):
        """Remove records which are older than limit."""
        ids = self.keys()
        while ids:
            r = ids.pop()
            rec = self[r]
            if not (("remove_stamp" in rec) or \
                    ("remove_id" in rec)):
                break
            elif len(self) <= 0:
                break
        for r in ids:
            rec = self[r]
            if rec.stamp + limit < now:
                rec.remove()

# End of Cache


class CacheList(list):

    """All cache."""

    def __init__(self):
        list.__init__(self)
        for i in os.listdir(config.cache_dir):
            c = Cache(i)
            self.append(c)

    def getall(self, timelimit=0):
        """Search nodes and update my cache."""
        random.shuffle(self)
        nodelist = NodeList()
        myself = nodelist.myself()
        searchlist = SearchList()
        for cache in self:
            if int(time.time()) > timelimit:
                sys.stderr.write("client timeout\n")
                return
            elif cache.size > config.file_limit*1024*1024:
                pass
            elif not cache.exists():
                pass
            else:
                node = searchlist.search(cache, myself=myself)
                if node is not None:
                    nodelist = NodeList()
                    if node not in nodelist:
                        nodelist.append(node)
                    cache.get_with_range(node)
            cache.sync_body()
            cache.check_attach()

    def search(self, query):
        result = []
        query = query.lower()
        for cache in self:
            f = open(cache.datpath + "/body.cache")
            buf = f.read()
            f.close()
            if buf.lower().find(query) >= 0:
                result.append(cache)
        return result

    def clean_records(self):
        """Clean old records."""
        now = int(time.time())
        for cache in self:
            if cache.save_record > 0:
                cache.remove_records(now, cache.save_record)

    def remove_removed(self):
        """Remove removed record from disk."""
        now = int(time.time())
        for cache in self:
            for r in os.listdir(cache.datpath + "/removed"):
                rec = Record(datfile=cache.datfile, idstr=r)
                if (cache.save_removed > 0) and \
                   (rec.stamp + cache.save_removed < now) and \
                   (rec.stamp < cache.stamp):
                    try:
                        os.remove(cache.datpath + "/removed/" + r)
                    except OSError, (errno, errorstr):
                        sys.stderr.write("OSError: %s: %s\n" %
                                         (cache.datpath + "/removed/" + r,
                                          errorstr))

# End of CacheList


class UpdateList(list):

    """Save update information.

    Get update if the system does not have the update.
    """

    update_file = config.update
    update_range = config.update_range

    def __init__(self):
        list.__init__(self)
        try:
            f = file(self.update_file)
            for line in f:
                buf = line.strip().split("<>")
                if len(buf) > 2:
                    idstr = buf[0] + "_" + buf[1]
                    r = Record(datfile=buf[2], idstr=idstr)
                    self.append(r)
            f.close()
        except IOError:
            sys.stderr.write(self.update_file + ": IOError\n")

    def sync(self):
        """Save update list."""
        f = file(self.update_file, "wb")
        now = int(time.time())
        for r in self:
            if now - self.update_range < r.stamp:
                f.write("<>".join((str(r.stamp), r.id, r.datfile)) + "\n")
        f.close()

# End of UpdateList
